@@ -13,7 +13,11 @@ xformToAffine <- function (image, useQuaternionFirst = TRUE)
         matrix <- diag(4)
         matrix[1:3,4] <- c(image@qoffset_x, image@qoffset_y, image@qoffset_z)
         q <- c(image@quatern_b, image@quatern_c, image@quatern_d)
-        q <- c(sqrt(1 - sum(q^2)), q)
+        
+        if (sum(q^2) == 1)
+            q <- c(0, q)
+        else
+            q <- c(sqrt(1 - sum(q^2)), q)
         
         matrix[1:3,1:3] <- c(q[1]*q[1] + q[2]*q[2] - q[3]*q[3] - q[4]*q[4],
                              2*q[2]*q[3] + 2*q[1]*q[4],
@@ -29,7 +33,7 @@ xformToAffine <- function (image, useQuaternionFirst = TRUE)
         # 0 should be treated as 1
         # This formulation does that (the 0.1 is arbitrary)
         qfactor <- sign(image@pixdim[1] + 0.1)
-        matrix[1:3,1:3] <- matrix[1:3,1:3] * c(abs(image@pixdim[2:3]), qfactor*abs(image@pixdim[4]))
+        matrix[1:3,1:3] <- matrix[1:3,1:3] * rep(c(abs(image@pixdim[2:3]), qfactor*abs(image@pixdim[4])), each=3)
         
         return (matrix)
     }
